import sys
from unittest import TestCase

from daudinlib.interaction import REPL
from daudinlib.pipeline import Pipeline


class TestREPL(TestCase):
    """Test the REPL class."""

    def testSetPrompts(self):
        """
        Setting the prompts via __init__ must work.
        """
        pl = Pipeline(loadInitFile=False)
        repl = REPL(pipeline=pl, ps1='x', ps2='y')
        self.assertEqual('x', repl.prompt)
        self.assertEqual('x', sys.ps1)
        self.assertEqual('y', sys.ps2)

    def testToggleDebug(self):
        """
        The REPL instance must be able to toggle the pipeline debug setting.
        """
        pl = Pipeline(loadInitFile=False)
        repl = REPL(pl)
        repl.runCommandLine('%d')
        self.assertIs(True, pl.debug)
        repl.runCommandLine('%d')
        self.assertIs(False, pl.debug)

    def testToggleTracebacks(self):
        """
        The REPL instance must be able to toggle the printing of tracebacks.
        """
        pl = Pipeline(loadInitFile=False)
        repl = REPL(pl)
        repl.runCommandLine('%t')
        self.assertIs(True, pl.printTracebacks)
        self.assertIs(True, pl.debug)
        repl.runCommandLine('%t')
        self.assertIs(False, pl.printTracebacks)
        # Note that turning off traceback printing does not turn off debug.
        self.assertIs(True, pl.debug)

    def testAttributes(self):
        """
        A REPL instance must have the expected attributes.
        """
        pl = Pipeline(loadInitFile=False)
        repl = REPL(pl)
        self.assertIs(pl, repl.pipeline)
        self.assertEqual(REPL.DEFAULT_PS1, repl.prompt)

    def testNumber(self):
        """
        The REPL instance must be in the expected state after processing a
        number.
        """
        pl = Pipeline(loadInitFile=False)
        repl = REPL(pl)
        repl.runCommandLine('4')
        self.assertAlmostEqual(4, pl.stdin)
        self.assertEqual(REPL.DEFAULT_PS1, repl.prompt)

    def testNumberPipeOneLine(self):
        """
        The REPL instance must be in the expected state after processing a
        command line containing a number followed by a pipe symbol.
        """
        pl = Pipeline(loadInitFile=False)
        repl = REPL(pl)
        repl.runCommandLine('4 |')
        self.assertAlmostEqual(4, pl.stdin)
        self.assertEqual(REPL.DEFAULT_PS1, repl.prompt)

    def testNumberPipeTwoLines(self):
        """
        The REPL instance must be in the expected state after processing a
        command line containing a number and then a second command line with
        just a pipe symbol.
        """
        pl = Pipeline(loadInitFile=False)
        repl = REPL(pl)
        repl.runCommandLine('4')
        self.assertFalse(pl.inPipeline)
        self.assertEqual(4, pl.stdin)
        repl.runCommandLine('')
        self.assertEqual(4, pl.stdin)
        self.assertEqual(REPL.DEFAULT_PS1, repl.prompt)

    def testIncompleteDict(self):
        """
        The REPL instance must be in the expected state after processing a
        an incomplete dict.
        """
        pl = Pipeline(loadInitFile=False)
        repl = REPL(pl)
        repl.runCommandLine('{ "a": 6, "b": ')
        # Is this next line correct? Should it be None?
        self.assertIs(None, pl.stdin)
        self.assertEqual(REPL.DEFAULT_PS2, repl.prompt)

    def testDictOnTwoLines(self):
        """
        The REPL instance must be in the expected state after processing a
        a dict given on two lines.
        """
        pl = Pipeline(loadInitFile=False)
        repl = REPL(pl)
        repl.runCommandLine('{ "a": 6, "b": ')
        repl.runCommandLine('7 }')
        self.assertEqual({'a': 6, 'b': 7}, pl.stdin)
        self.assertEqual(REPL.DEFAULT_PS1, repl.prompt)

    def testInitialEmptyCommands(self):
        """
        We should be able to hit ENTER repeatedly in a new session, while
        stdin remains as None and the prompt is unchanged.
        """
        pl = Pipeline(loadInitFile=False)
        repl = REPL(pl)

        repl.runCommandLine('')
        self.assertEqual(None, pl.stdin)
        self.assertEqual(REPL.DEFAULT_PS1, repl.prompt)

        repl.runCommandLine('')
        self.assertEqual(None, pl.stdin)
        self.assertEqual(REPL.DEFAULT_PS1, repl.prompt)

    def testInitialWhitespaceCommands(self):
        """
        We should be able to repeatedly enter whitespace commands in a new
        session, while stdin remains as None and the prompt is unchanged.
        """
        pl = Pipeline(loadInitFile=False)
        repl = REPL(pl)

        repl.runCommandLine('  ')
        self.assertEqual(None, pl.stdin)
        self.assertEqual(REPL.DEFAULT_PS1, repl.prompt)

        repl.runCommandLine('  \t  ')
        self.assertEqual(None, pl.stdin)
        self.assertEqual(REPL.DEFAULT_PS1, repl.prompt)

    def testEmptyCommands(self):
        """
        We should be able to hit ENTER repeatedly without changing
        the pending stdin value or the prompt.
        """
        pl = Pipeline(loadInitFile=False)
        repl = REPL(pl)

        repl.runCommandLine('4')
        self.assertEqual(4, pl.stdin)
        self.assertEqual(REPL.DEFAULT_PS1, repl.prompt)

        repl.runCommandLine('')
        self.assertEqual(4, pl.stdin)
        self.assertEqual(REPL.DEFAULT_PS1, repl.prompt)

        repl.runCommandLine('')
        self.assertEqual(4, pl.stdin)
        self.assertEqual(REPL.DEFAULT_PS1, repl.prompt)

    def testWhitespaceCommands(self):
        """
        We should be able to repeatedly enter whitespace commands without
        changing the pending stdin value or the prompt.
        """
        pl = Pipeline(loadInitFile=False)
        repl = REPL(pl)

        repl.runCommandLine('4')
        self.assertEqual(4, pl.stdin)
        self.assertEqual(REPL.DEFAULT_PS1, repl.prompt)

        repl.runCommandLine(' ')
        self.assertEqual(4, pl.stdin)
        self.assertEqual(REPL.DEFAULT_PS1, repl.prompt)

        repl.runCommandLine('\t\t')
        self.assertEqual(4, pl.stdin)
        self.assertEqual(REPL.DEFAULT_PS1, repl.prompt)

    def testEchoCat(self):
        """
        We should be able to pipe echo output to cat.
        """
        pl = Pipeline(loadInitFile=False)
        repl = REPL(pl)
        repl.runCommandLine('echo hi | cat')
        self.assertEqual(['hi'], pl.stdin)
        self.assertEqual(REPL.DEFAULT_PS1, repl.prompt)

    def testEchoCatCat(self):
        """
        We should be able to pipe echo output to cat twice.
        """
        pl = Pipeline(loadInitFile=False)
        repl = REPL(pl)
        repl.runCommandLine('echo hi | cat | cat')
        self.assertEqual(['hi'], pl.stdin)
        self.assertEqual(REPL.DEFAULT_PS1, repl.prompt)

    def testAddUnderscoreVar(self):
        """
        We should be able to add to the _ variable even when not in
        a pipeline.
        """
        pl = Pipeline(loadInitFile=False)
        repl = REPL(pl)
        repl.runCommandLine('7')
        repl.runCommandLine('_ + 10')
        self.assertEqual(17, pl.stdin)
        self.assertEqual(REPL.DEFAULT_PS1, repl.prompt)


class TestREADME(TestCase):
    """Test the examples from the README."""

    def testAreaFunction(self):
        """Define an area function and call it."""
        pl = Pipeline(loadInitFile=False)
        repl = REPL(pl)

        repl.runCommandLine('from math import pi')
        self.assertEqual(REPL.DEFAULT_PS1, repl.prompt)

        repl.runCommandLine('def area(r):')
        self.assertEqual(REPL.DEFAULT_PS2, repl.prompt)

        repl.runCommandLine('  return r ** 2 * pi')
        self.assertEqual(REPL.DEFAULT_PS2, repl.prompt)

        repl.runCommandLine('')
        self.assertEqual(REPL.DEFAULT_PS1, repl.prompt)

        repl.runCommandLine('area(2.0)')
        self.assertEqual(REPL.DEFAULT_PS1, repl.prompt)

        self.assertAlmostEqual(12.566370614359172, pl.stdin)

    def testTripleFunction(self):
        pl = Pipeline(loadInitFile=False)
        repl = REPL(pl)

        repl.runCommandLine('def triple(x):')
        self.assertEqual(REPL.DEFAULT_PS2, repl.prompt)

        repl.runCommandLine('  return int(x) * 3')
        self.assertEqual(REPL.DEFAULT_PS2, repl.prompt)

        repl.runCommandLine('')
        self.assertEqual(REPL.DEFAULT_PS1, repl.prompt)

        repl.runCommandLine('echo a b c | wc -w | triple(_[0])')
        self.assertEqual(REPL.DEFAULT_PS1, repl.prompt)

        self.assertEqual(9, pl.stdin)

    def testTripleFunctionWithLinebreaks(self):
        pl = Pipeline(loadInitFile=False)
        repl = REPL(pl)

        repl.runCommandLine('def triple(x):')
        self.assertEqual(REPL.DEFAULT_PS2, repl.prompt)

        repl.runCommandLine('  return int(x) * 3')
        self.assertEqual(REPL.DEFAULT_PS2, repl.prompt)

        repl.runCommandLine('')
        self.assertEqual(REPL.DEFAULT_PS1, repl.prompt)

        repl.runCommandLine('echo a b c')
        self.assertEqual(REPL.DEFAULT_PS1, repl.prompt)

        repl.runCommandLine('| wc -w |')
        self.assertEqual(REPL.DEFAULT_PS1, repl.prompt)

        repl.runCommandLine('triple(_[0])')
        self.assertEqual(REPL.DEFAULT_PS1, repl.prompt)

        self.assertEqual(9, pl.stdin)

    def testTripleFunctionWithIntermediateCleanup(self):
        pl = Pipeline(loadInitFile=False)
        repl = REPL(pl)

        repl.runCommandLine('def triple(x):')
        self.assertEqual(REPL.DEFAULT_PS2, repl.prompt)

        repl.runCommandLine('  return int(x) * 3')
        self.assertEqual(REPL.DEFAULT_PS2, repl.prompt)

        repl.runCommandLine('')
        self.assertEqual(REPL.DEFAULT_PS1, repl.prompt)

        repl.runCommandLine('echo a b c | wc -w |')
        self.assertEqual(REPL.DEFAULT_PS1, repl.prompt)

        pl.toggleDebug()
        repl.runCommandLine('ll = lambda x: x[0]')
        self.assertEqual(REPL.DEFAULT_PS1, repl.prompt)

        repl.runCommandLine('ll(_)')
        self.assertEqual(REPL.DEFAULT_PS1, repl.prompt)

        repl.runCommandLine('triple(_)')
        self.assertEqual(REPL.DEFAULT_PS1, repl.prompt)

        self.assertEqual(9, pl.stdin)

    def testAbsArithmetic(self):
        pl = Pipeline(loadInitFile=False)
        repl = REPL(pl)
        repl.runCommandLine('-6 | abs(_) | _ * 7')
        self.assertEqual(REPL.DEFAULT_PS1, repl.prompt)
        self.assertEqual(42, pl.stdin)

    def testInterpretingSection(self):
        """
        Test the commands carried out in the 'How commands are interpreted'
        section.
        """
        pl = Pipeline(loadInitFile=False)
        repl = REPL(pl)
        repl.runCommandLine('4')
        self.assertEqual(4, pl.stdin)

        repl.runCommandLine('_')
        self.assertEqual(4, pl.stdin)

        repl.runCommandLine('[3, 6, 9]')
        self.assertEqual([3, 6, 9], pl.stdin)

        repl.runCommandLine("print('hello')")
        self.assertEqual('hello', pl.stdin)

        repl.runCommandLine('echo hello too')
        self.assertEqual(['hello too'], pl.stdin)

        self.assertEqual(REPL.DEFAULT_PS1, repl.prompt)



import os
import sys
import re
import select
import termios
import tty
import pty
from code import compile_command
from io import StringIO, TextIOWrapper
from contextlib import contextmanager
from os.path import exists, join, expanduser
import traceback
from subprocess import Popen, PIPE, CalledProcessError, run

_originalStdout = sys.stdout

# The following escape sequence regex is taken from
# https://stackoverflow.com/questions/14693701/\
# how-can-i-remove-the-ansi-escape-sequences-from-a-string-in-python
# Answer by https://stackoverflow.com/users/100297/martijn-pieters
ANSI_esc = re.compile(r'''
    \x1B    # ESC
    [@-_]   # 7-bit C1 Fe
    [0-?]*  # Parameter bytes
    [ -/]*  # Intermediate bytes
    [@-~]   # Final byte
''', re.VERBOSE)


@contextmanager
def newStdout(stdout=None):
    originalStdout = sys.stdout
    stdout = stdout or StringIO()
    sys.stdout = stdout
    try:
        yield stdout
    except Exception:
        raise
    finally:
        sys.stdout = originalStdout


@contextmanager
def newStdin(fp):
    originalStdin = sys.stdin
    sys.stdin = fp
    try:
        yield fp
    except Exception:
        raise
    finally:
        sys.stdin = originalStdin


class Pipeline:

    IGNORE = object()

    def __init__(self, outfp=sys.stdout, errfp=sys.stderr, loadInitFile=True):
        self.outfp = outfp
        self.errfp = errfp
        self.stdin = None
        self.lastStdin = None
        self.stdout = None
        self.pendingText = ''
        self.debug = False
        self.printTracebacks = False
        self.initFile = join(expanduser('~'), '.daudin.py')
        self.lastResultIsList = None
        self.local = {}
        if loadInitFile:
            self.loadInitFile()
        self.initializeLocal()
        self.inPipeline = False

    @property
    def incomplete(self):
        return bool(self.pendingText)

    def loadInitFile(self):
        """
        Load the user's initialization file.

        @return: A C{bool} indicating whether an init file was loaded.
        """
        if exists(self.initFile):
            exec(open(self.initFile).read(), self.local)
            return True
        return False

    def initializeLocal(self):
        """Update the C{self.local} dict to be used with eval/exec."""
        for key, value in (('self', self), ('sh', self.sh), ('cd', self.cd)):
            self.local.setdefault(key, value)

    def run(self, command, commandNumber=1, nCommands=1):
        self._debug('Processing %r.' % command)
        self.lastStdin = self.stdin
        self.lastResultIsList = False
        strippedCommand = command.strip()

        if self.pendingText:
            # The previous command was incomplete.
            fullCommand = self.pendingText + '\n' + (
                command if strippedCommand else '')
        else:
            fullCommand = strippedCommand

        # This command is executing as part of a pipeline if:
        #   1) we were already in a pipeline (due to the last command line) or
        #   2) this is not the 1st command of a multi-command command line, or
        #   3) this is the first command  of a multi-command command line but
        #      the command is equal (this happens when a command line starts
        #      with a pipe symbol).
        self.inPipeline = self.inPipeline or commandNumber > 1 or (
            nCommands > 1 and commandNumber == 1 and not fullCommand)

        self._debug('%s pipeline.' % ('In' if self.inPipeline else 'Not in'))

        print_ = (commandNumber == nCommands)

        handled, doPrint = self._tryEval(fullCommand, print_)

        if not handled:
            handled, doPrint = self._tryExec(fullCommand, print_)

        if not handled:
            handled, doPrint = self._tryShell(fullCommand, print_)

        if handled:
            if commandNumber == nCommands:
                self.inPipeline = not fullCommand
        else:
            print('Could not handle command %r' % command, file=self.errfp)
            self.reset()

        return bool(self.pendingText), doPrint

    def _tryEval(self, strippedCommand, print_):
        if not strippedCommand:
            self._debug('Eval skipped (command empty).')
            return False, False

        self._debug('Trying eval %r.' % (strippedCommand,))
        try:
            with newStdout() as so:
                self.local['_'] = self.stdin
                result = eval(strippedCommand, self.local)
        except Exception as e:
            self._debug('Could not eval: %s.' % e)
            if self.printTracebacks:
                self._debug(traceback.format_exc())
            return False, False
        else:
            self._debug('Eval returned %r.' % (result,))
            self.pendingText = ''
            if isinstance(result, str):
                if result.endswith('\n'):
                    result = result[:-1]
            elif result is None:
                stdout = so.getvalue()
                if stdout:
                    self._debug('Eval printed %r.' % (stdout,))
                    if stdout.endswith('\n'):
                        stdout = stdout[:-1]
                    if stdout.find('\n') > -1:
                        result = stdout.split('\n')
                        self.lastResultIsList = True
                    else:
                        result = stdout
                else:
                    print_ = False

            if result is self.IGNORE:
                print_ = False
            else:
                self.stdin = result

            return True, print_

    def _tryExec(self, command, print_):
        self._debug('Trying to compile %r.' % (command,))

        exception = None

        try:
            codeobj = compile_command(command)
        except (OverflowError, SyntaxError, ValueError) as e:
            self._debug('%s: %s.' % (e.__class__.__name__, e))
            if self.printTracebacks:
                self._debug(traceback.format_exc())
            self.pendingText = ''
            exception = e
        else:
            self._debug('Command compiled OK.')
            so = StringIO()
            if codeobj:
                self.local['_'] = self.stdin
                with newStdout(so):
                    try:
                        exec(codeobj, self.local)
                    except Exception as e:
                        self._debug('Could not exec: %s.' % e)
                        if self.printTracebacks:
                            self._debug(traceback.format_exc())
                        exception = e
                    else:
                        self._debug('Exec succeeded.')
                self.pendingText = ''
            else:
                self._debug('Incomplete command.')
                self.pendingText = command

        if exception is None:
            if self.pendingText:
                print_ = False
            else:
                stdout = so.getvalue()
                if stdout:
                    self._debug('Exec printed %r.' % (stdout,))
                    if stdout.endswith('\n'):
                        stdout = stdout[:-1]
                    if stdout.find('\n') > -1:
                        self.stdin = stdout.split('\n')
                        self.lastResultIsList = True
                    else:
                        self.stdin = stdout
                else:
                    print_ = False

            return True, print_
        else:
            return False, False

    def _tryShell(self, command, print_):
        self._debug('Trying shell %r with stdin %r.' % (command, self.stdin,))
        try:
            result = self.sh(command, print_=print_)
        except CalledProcessError as e:
            print('Process error: %s' % e, file=sys.errfp)
            return False, False

        self._debug('Shell returned %r' % (result,))
        if result:
            if result.endswith('\n'):
                result = result[:-1]
            self.stdin = result.split('\n')
            # Set lastResultIsList to False because the result has
            # already been printed in its non-list form. So next time
            # we print it we want to see the list.
            self.lastResultIsList = False
        else:
            self.stdin = []

        return True, False

    def sh(self, *args, print_=False, **kwargs):
        """
        Execute a shell command, with input from our C{self.stdin}.

        @param args: Positional arguments to pass to C{subprocess.run}.
        @param print_: If C{True}, use a pseudo-tty and print the output to
            stdout.
        @param kwargs: Keyword arguments to pass to C{Pipe} or
            C{subprocess.run} (depending on the value of C{print_}).
        @raise CalledProcessError: If the command results in an error.
        @return: The C{str} output of the command.
        """
        kwargs.setdefault('shell', len(args) == 1 and isinstance(args[0], str))

        if self.inPipeline:
            if self.stdin is None:
                stdin = None
            elif isinstance(self.stdin, list):
                stdin = '\n'.join(map(str, self.stdin)) + '\n'
            else:
                stdin = str(self.stdin) + '\n'
        else:
            stdin = None

        run = self._shPty if print_ else self._sh

        return run(stdin, *args, **kwargs)

    def _sh(self, stdin, *args, **kwargs):
        """
        Execute a shell command, with input from C{stdin}.

        @param stdin: The C{str} input to the process, else C{None}.
        @param args: Positional arguments to pass to C{subprocess.run}.
        @param kwargs: Keyword arguments to pass to C{subprocess.run}.
        @raise CalledProcessError: If the command results in an error.
        @return: The C{str} output of the command.
        """
        self._debug('In _sh, stdin is %r' % (stdin,))
        kwargs.setdefault('input', stdin)
        kwargs.setdefault('stdout', PIPE)
        kwargs.setdefault('universal_newlines', True)
        return run(*args, **kwargs).stdout

    def _shPty(self, stdin, *args, **kwargs):
        """
        Run a command in a pseudo-tty, with input from C{stdin}.
        """
        self._debug('In _shPty, stdin is %r' % (stdin,))

        # Detect if we're running under pytest, in which case stdin cannot
        # be used.
        pytest = 'pytest' in sys.modules

        # The following is (slightly) adapted from
        # https://stackoverflow.com/questions/41542960/\
        # run-interactive-bash-with-popen-and-a-dedicated-tty-python
        # Answer by https://stackoverflow.com/users/3555925/liao

        # save original tty setting then set it to raw mode
        if not pytest:
            oldTty = termios.tcgetattr(sys.stdin)

        try:
            if not pytest:
                tty.setraw(sys.stdin.fileno())

            # Open a pseudo-terminal to interact with the subprocess.
            master_fd, slave_fd = pty.openpty()

            # Pass os.setsid to have the process run in a new process group.
            #
            # Note that we should be more careful with kwargs here. If the
            # user has called 'sh' interactively and passed a value for
            # stdin, stdout, stderr, or universal_newlines, the following
            # will cause Python to complain about multiple values for a
            # keyword argument. We should check & warn the user etc.
            process = Popen(
                *args, preexec_fn=os.setsid,
                stdin=(slave_fd if stdin is None else PIPE), stdout=slave_fd,
                stderr=slave_fd, universal_newlines=True, **kwargs)

            # Write the command's stdin to it, if any.
            if stdin is not None:
                # print('WROTE %r' % (stdin,), file=self.errfp)
                os.write(process.stdin.fileno(), stdin.encode())
                process.stdin.close()

            if pytest:
                readFds = [master_fd]
            else:
                readFds = [sys.stdin, master_fd]

            result = b''
            while process.poll() is None:
                r, w, e = select.select(readFds, [], [], 0.05)
                if sys.stdin in r:
                    data = os.read(sys.stdin.fileno(), 10240)
                    # print('READ from stdin %r' % (data,), file=self.errfp)
                    if data == b'\x03':
                        process.terminate()
                    else:
                        os.write(master_fd, data)
                elif master_fd in r:
                    data = os.read(master_fd, 10240)
                    # print('READ from master %r' % (data,), file=self.errfp)
                    if data:
                        result += data
                        os.write(_originalStdout.fileno(), data)

        finally:
            if not pytest:
                # Restore tty settings.
                termios.tcsetattr(sys.stdin, termios.TCSADRAIN, oldTty)

        # TODO: Check all this still needed.
        # print('Shell result %r' % (result,), file=self.errfp)
        return ANSI_esc.sub('', result.decode('utf-8')).replace('\r\n', '\n')

    def cd(self, dest):
        os.chdir(dest)
        return self.IGNORE

    def toggleDebug(self):
        self.debug = not self.debug

    def toggleTracebacks(self):
        self.printTracebacks = not self.printTracebacks

    def undo(self):
        self.stdin = self.lastStdin

    def reset(self):
        self.lastStdin = self.stdin
        self.stdin = None
        self.pendingText = ''
        self.lastResultIsList = None
        self.inPipeline = False

    def _debug(self, *args, **kwargs):
        if self.debug:
            kwargs.setdefault('file', self.errfp)
            origEnd = kwargs.get('end', '\n')
            kwargs['end'] = ''
            print(' ' * 20, **kwargs)
            kwargs['end'] = origEnd
            print(*args, **kwargs)
        return self.IGNORE

    def print_(self):
        if isinstance(self.stdin, TextIOWrapper):
            s = self.stdin.read()
            print(s, end='' if s.endswith('\n') else '\n', file=self.outfp)
        elif isinstance(self.stdin, str):
            print(self.stdin, end='' if self.stdin.endswith('\n') else '\n',
                  file=self.outfp)
        elif self.lastResultIsList:
            print('\n'.join(self.stdin), file=self.outfp)
        else:
            print(self.stdin, file=self.outfp)



import re

_unescapedPipeRegex = re.compile(r'(?<!\\)\|')


def lineSplitter(line):
    start = 0
    for match in _unescapedPipeRegex.finditer(line):
        yield line[start:match.start()].replace(r'\|', '|')
        start = match.end()

    yield line[start:].replace(r'\|', '|')



import readline
import os
import glob


class Completor():

    def complete(self, text, state):
        if state == 0:
            self.completions = []
            for path in glob.glob(text + '*'):
                if os.path.isdir(path):
                    if not path.endswith(os.sep):
                        path += os.sep
                else:
                    path += ' '
                self.completions.append(path)

        try:
            return self.completions[state]
        except IndexError:
            return None


def setupReadline():
    """Initialize the readline library and command history.

    @return: A C{bool} to indicate whether standard input is a terminal
        (and therefore interactive).
    """
    if not os.isatty(0):
        # Standard input is closed or is a pipe etc. So there's no user
        # typing at us, and so no point in setting up readline.
        return False

    readline.parse_and_bind('tab: complete')
    readline.set_completer_delims(' \t\n')
    readline.set_completer(Completor().complete)

    # Readline code from https://docs.python.org/3.7/library/readline.html
    histfile = os.path.join(os.path.expanduser('~'), '.daudin_history')

    try:
        readline.read_history_file(histfile)
        historyLen = readline.get_current_history_length()
    except FileNotFoundError:
        open(histfile, 'wb').close()
        historyLen = 0

    try:
        readline.append_history_file
    except AttributeError:
        # We won't be able to save readline history. This can happen on
        # Python 3.5 at least - not sure why.
        pass
    else:
        import atexit

        def saveHistory(prevHistoryLen, histfile):
            newHistoryLen = readline.get_current_history_length()
            readline.set_history_length(1000)
            readline.append_history_file(newHistoryLen - prevHistoryLen,
                                         histfile)

        atexit.register(saveHistory, historyLen, histfile)

    return True



from unittest import TestCase

from daudinlib.pipeline import Pipeline


class TestPipeline(TestCase):
    """Test the Pipeline class."""

    def testNumber(self):
        """A number should be processed correctly."""
        p = Pipeline(loadInitFile=False)
        p.run('4')
        self.assertEqual(4, p.stdin)

    def testString(self):
        """A string should be processed correctly."""
        p = Pipeline(loadInitFile=False)
        p.run('"hello"')
        self.assertEqual('hello', p.stdin)

    def testTrue(self):
        """True should be processed correctly."""
        p = Pipeline(loadInitFile=False)
        p.run('True')
        self.assertIs(True, p.stdin)

    def testFalse(self):
        """False should be processed correctly."""
        p = Pipeline(loadInitFile=False)
        p.run('False')
        self.assertIs(False, p.stdin)

    def testNone(self):
        """None should be processed correctly."""
        p = Pipeline(loadInitFile=False)
        p.run('None')
        self.assertEqual(None, p.stdin)

    def testArithmetic(self):
        """Arithmetic should work."""
        p = Pipeline(loadInitFile=False)
        p.run('(3 + 4 + 5) / 4')
        self.assertEqual(3, p.stdin)

    def testFloats(self):
        """Floating point arithmetic should work."""
        p = Pipeline(loadInitFile=False)
        p.run('6.3 / 1.8')
        self.assertEqual(3.5, p.stdin)

    def testInPipelineWithTrailingPipe(self):
        """
        The pipeline must have inPipeline True when the command line ends in a
        pipe.
        """
        p = Pipeline(loadInitFile=False)
        p.run('6', 1, 2)
        p.run('', 2, 2)
        self.assertTrue(p.inPipeline)
        p.run('6', 1, 2)
        self.assertTrue(p.inPipeline)

    def testNotInPipelineInitially(self):
        """
        The pipeline must have inPipeline False when it is created.
        """
        p = Pipeline(loadInitFile=False)
        self.assertFalse(p.inPipeline)

    def testInPipelineWithLeadingPipe(self):
        """
        The pipeline must have inPipeline True when the command line starts
        with a pipe.
        """
        p = Pipeline(loadInitFile=False)
        p.run('6', 1, 1)
        p.run('', 1, 2)
        self.assertTrue(p.inPipeline)

    def testInPipelineWithIntermediateCommand(self):
        """
        The pipeline must have inPipeline True when the command is one of the
        middle commands of a multi-command command line.
        """
        p = Pipeline(loadInitFile=False)
        p.run('6', 3, 5)
        self.assertTrue(p.inPipeline)


class TestREADME(TestCase):
    """Test the examples from the README."""

    def testAreaFunction(self):
        """Define an area function and call it."""
        p = Pipeline(loadInitFile=False)
        p.run('from math import pi')
        self.assertFalse(p.incomplete)
        # self.assertFalse(p.inPipeline)

        p.run('def area(r):')
        self.assertTrue(p.incomplete)
        # self.assertFalse(p.inPipeline)

        p.run('  return r ** 2 * pi', True)
        self.assertTrue(p.incomplete)
        # self.assertFalse(p.inPipeline)

        p.run('', True)
        self.assertFalse(p.incomplete)
        # self.assertFalse(p.inPipeline)

        p.run('area(2.0)')
        self.assertFalse(p.incomplete)
        # self.assertFalse(p.inPipeline)

        self.assertAlmostEqual(12.566370614359172, p.stdin)



from unittest import TestCase

from daudinlib.parse import lineSplitter


class TestLineSplitter(TestCase):
    """Test the lineSplitter function."""

    def testEmptyString(self):
        """An empty string should result in a list with an empty string."""
        self.assertEqual([''],
                         list(lineSplitter('')))

    def testString(self):
        """A string with no | should result in the same string."""
        self.assertEqual(['hello'],
                         list(lineSplitter('hello')))

    def testPlainStrings(self):
        """An unescaped | should result in the two expected fields."""
        self.assertEqual(['hello ', ' there'],
                         list(lineSplitter('hello | there')))

    def testEscaped(self):
        """An escaped | should result in one field, with the escape remvoed."""
        self.assertEqual([r'echo hi | wc -c'],
                         list(lineSplitter(r'echo hi \| wc -c')))



from __future__ import print_function

import sys
import traceback
from os.path import expanduser

from daudinlib.parse import lineSplitter
from daudinlib.pipeline import Pipeline


class REPL():

    DEFAULT_PS1 = '>>> '
    DEFAULT_PS2 = '... '

    def __init__(self, pipeline=None, ps1=DEFAULT_PS1, ps2=DEFAULT_PS2):
        self.pipeline = pipeline or Pipeline()
        sys.ps1 = self.prompt = ps1
        sys.ps2 = ps2

    def reset(self):
        self.pipeline.reset()
        self.prompt = sys.ps1

    def _readStdin(self):
        while True:
            try:
                text = input(self.prompt)
            except KeyboardInterrupt:
                print('^C', file=sys.stderr)
                self.reset()
            except EOFError:
                print('^D', file=sys.stderr)
                sys.exit()
            else:
                yield text

    def interact(self):
        for commandLine in self._readStdin():
            self.runCommandLine(commandLine)

    def runCommandLine(self, text):
        commands = list(lineSplitter(text))
        nCommands = len(commands)
        for i, command in enumerate(commands, start=1):
            if self.runCommand(command, i, nCommands) is False:
                return False
        return True

    def runCommand(self, command, commandNumber=1, nCommands=1):
        pipeline = self.pipeline

        if self._handleSpecial(command):
            return

        try:
            incomplete, doPrint = pipeline.run(command, commandNumber,
                                               nCommands)
        except KeyboardInterrupt:
            print('^C', file=sys.stderr)
            self.reset()
            return False
        except Exception:
            print(traceback.format_exc(), file=sys.stderr)
            self.reset()
            return False
        else:
            self.prompt = sys.ps2 if incomplete else sys.ps1
            if doPrint:
                pipeline.print_()

        return True

    def _handleSpecial(self, command):
        strippedCommand = command.strip()
        pipeline = self.pipeline

        if strippedCommand.startswith('%cd'):
            if strippedCommand == '%cd':
                dir_ = expanduser('~')
            else:
                dir_ = command.lstrip().split(None, 1)[1]
            try:
                pipeline.cd(dir_)
            except FileNotFoundError:
                print('No such directory %r' % dir_, file=sys.stderr)
            return True

        if strippedCommand == '%d':
            pipeline.toggleDebug()
            return True

        if strippedCommand == '%r':
            if pipeline.loadInitFile():
                print('Reloaded.', file=sys.stderr)
            else:
                print('Daudin init file %r does not exist.' %
                      pipeline.initFile, file=sys.stderr)
            return True

        if strippedCommand == '%t':
            pipeline.toggleTracebacks()
            # If we just turned on traceback printing, make sure debug
            # printing is also on.
            if pipeline.printTracebacks:
                pipeline.debug = True
            return True

        if strippedCommand == '%u':
            pipeline.undo()
            return True

        if strippedCommand == '_':
            pipeline.print_()
            return True

        return False



#!/usr/bin/env python

from setuptools import setup


# Modified from http://stackoverflow.com/questions/2058802/
# how-can-i-get-the-version-defined-in-setup-py-setuptools-in-my-package and
# https://stackoverflow.com/questions/6786555/
# automatic-version-number-both-in-setup-py-setuptools-and-source-code#7502821

def version():
    import os
    import re

    init = os.path.join(
        os.path.abspath(os.path.dirname(__file__)), 'daudinlib', '__init__.py')
    with open(init) as fp:
        initData = fp.read()
    match = re.search(r"^__version__ = ['\"]([^'\"]+)['\"]",
                      initData, re.M)
    if match:
        return match.group(1)
    else:
        raise RuntimeError('Unable to find version string in %r.' % init)


setup(name='daudin',
      version=version(),
      packages=['daudinlib'],
      include_package_data=False,
      url='https://github.com/terrycojones/daudin',
      download_url='https://github.com/terrycojones/daudin',
      author='Terry Jones',
      author_email='terry@jon.es',
      keywords=['python shell'],
      classifiers=[
          'Development Status :: 3 - Alpha',
          'Environment :: Console',
          'Programming Language :: Python :: 3 :: Only',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: MIT License',
          'Operating System :: MacOS :: MacOS X',
          'Operating System :: POSIX :: Linux',
          'Operating System :: Unix',
          'Topic :: System :: Shells'
      ],
      license='MIT',
      scripts=['daudin'],
      description='A UNIX command-line shell based on Python',
      extras_require={
        'dev': [
            'flake8',
            'pytest',
        ]
      })
